(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_def.mli 317 2010-11-01 17:44:31Z gerd $ *)

(** Defining the map/reduce job *)

class type mapred_env =
object
  method open_cluster : Unixqueue.event_system -> Plasma_client.plasma_cluster
    (** Opens the cluster filesystem *)

  method config : Mapred_config.mapred_config
    (** The map/reduce server configuration *)

  method config_file : Netplex_types.config_file
    (** The config file containing [config] *)
end


class type mapred_job_config =
object
  method name : string
    (** A name for identifying this job, e.g. in log files *)

  method job_id : string
    (** An automatically generated name. This can be considered as
	unique. It is not possible to override this name.
     *)

  method input_dir : string
    (** This plasma directory contains the input files *)

  method output_dir : string
    (** This plasma directory will get output files. It should exist,
	and it should be empty
     *)

  method work_dir : string
    (** This plasma directory is used for temporary files. It should
	exist, and it should be empty
     *)

  method log_dir : string
    (** This plasma directory is used for log files. It should exist,
	and it should be empty
     *)

  method task_files : string list
    (** These files are copied at job start to the "local" directory
	for this job on all task nodes. This should be regular files
	only.
     *)

  method bigblock_size : int
    (** Map/reduce processes files in units of bigblocks. The size of
	bigblocks can be chosen as multiples of the size of filesystem
	blocks. The value [bigblock_size] is in bytes. The maximum
	size of records (line length) is also [bigblock_size]. Reasonable
	values are in the multi-megabyte range, e.g. 16M.

	This value must not exceed [mr_buffer_size] or [mr_buffer_size_tight].
     *)

  method map_tasks : int
    (** The number of map tasks that should not be exceeded. It is tried
	to hit this number, but it may be possible that not enough map
	tasks can be generated.
     *)

  method merge_limit : int
    (** How many files are merged at most by a shuffle task *)

  method split_limit : int
    (** How many files are created by a shuffle task. This is not a
	strict limit - actually the scheduler plans with slightly
	larger split limits near the end of the job.  
     *)

  method partitions : int
    (** The number of partitions = number of reduce tasks *)

  method custom : string -> string
    (** Get a custom parameter or raise [Not_found] *)
end


class type task_info =
object
  method req_id : int
    (** The request ID. This is different for each task of a job *)

  method task_prefix : string
    (** This filename prefix is used for log files *)

  method partition : int
    (** The partition. Only available at reduce time *)
end


class type mapred_job =
object
  method custom_params : string list
    (** The list of allowed custom parameters *)

  method check_config : mapred_env -> mapred_job_config -> unit
    (** Check the config. If not ok, this method can raise exceptions to
	stop everything
     *)

  method pre_job_start : mapred_env -> mapred_job_config -> unit
    (** This is run by the job process before the first task is started *)

  method post_job_finish : mapred_env -> mapred_job_config -> unit
    (** This is run by the job process after the last task is finished *)

  method map : mapred_env -> mapred_job_config ->
               task_info -> Mapred_io.record_reader ->
                 Mapred_io.record_writer -> unit
    (** The mapper reads records, maps them, and writes them into a
	second file.
     *)

  method extract_key : mapred_env -> mapred_job_config -> string -> string
    (** Extracts the key from a record. This method is always called by
	first evaluating [let f = job#extract_key me jc], and then
	calling [f line] for each input line. Because of this, it is
	possible to factor initializations out as in

	{[
	   method extract_key me jc =
              ...; (* init stuff *)
              (fun line -> ...  (* real extraction *) )
	]}

 *)

  method partition_of_key : mapred_env -> mapred_job_config -> string -> int
    (** Determines the partition of a key. Can be something simple like
	[fun k -> (Hashtbl.hash k) mod partitions], or something more
	elaborated. This method is always called by
	first evaluating [let f = job#partition_of_key me jc], and then
	calling [f line] for each input line. Because of this, it is
	possible to factor initializations out as in

	{[
	   method partition_of_key me jc =
              ...; (* init stuff *)
              (fun line -> ...  (* real extraction *) )
	]}

     *)

  method reduce : mapred_env ->  mapred_job_config ->
                 task_info -> Mapred_io.record_reader -> 
                    Mapred_io.record_writer -> unit
    (** The reducer reads all the records of one partition, and puts them
	into an output file.
     *)

end
