(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: mapred_job_exec.ml 359 2011-01-31 15:46:09Z gerd $ *)

(** Where to run which task:
   - map task: is specified in the task
   - sort task: same node as preceding map
   - first round of shuffling: same node as preceding sort
   - further rounds of shuffling: any node
   - reduce task: same node as preceding shuffling task if this task
     produced only one file; otherwise any node
 *)

open Mapred_tasks
open Mapred_def
open Printf

module Mr_proxy = Mapred_rpc_clnt.Make'Mapred_task(Rpc_proxy.ManagedClient)

module TaskSet = Set.Make(Mapred_tasks.Ord)


type runtime_job_config =
    { map_weight : float * float;
      sort_weight : float * float;
      shuffle_weight : float * float;
      reduce_weight : float * float;
      load_limit : float;
      dump_plan_when_complete : bool;
      res_conf : Mapred_rpc_aux.resource_config;
      mjc : Mapred_job_config.m_job_config;
    }

let dlogf = Plasma_util.dlogf

let create_runtime_job_config 
      ?(map_weight=(0.9, 0.1))     (* (io, cpu) *) 
      ?(sort_weight=(0.1, 0.9)) 
      ?(shuffle_weight=(0.9, 0.1))
      ?(reduce_weight=(0.9, 0.5)) 
      ?(dump_plan_when_complete=false)
      ?shm_low
      ?shm_high
      ?shm_max
      ?buf_low
      ?buf_high
      ?buf_max
      ?(shm_low_factor=0.25)
      ?(shm_high_factor=0.5)
      ?(shm_max_factor=0.75)
      ?(buf_low_factor=0.25)
      ?(buf_high_factor=0.5)
      ?(buf_max_factor=0.75)
      ~load_limit mjc =
  let res_conf =
    { Mapred_rpc_aux.rc_shm_low = shm_low;
      rc_shm_high = shm_high;
      rc_shm_max = shm_max;
      rc_buf_low = buf_low;
      rc_buf_high = buf_high;
      rc_buf_max = buf_max;
      rc_shm_low_p = shm_low_factor;
      rc_shm_high_p = shm_high_factor;
      rc_shm_max_p = shm_max_factor;
      rc_buf_low_p = buf_low_factor;
      rc_buf_high_p = buf_high_factor;
      rc_buf_max_p = buf_max_factor;
    } in
  { map_weight;
    sort_weight;
    shuffle_weight;
    reduce_weight;
    load_limit;
    dump_plan_when_complete;
    mjc = mjc;
    res_conf;
  }

module MCSet = Set.Make(Rpc_proxy.ManagedClient)

type status =
    [ `Running | `Successful | `Errors of string list | `Killed ]

type running_job =
    { config : runtime_job_config;
      plan : Mapred_sched.plan;
      mutable started_tasks : task list;  (* for cleanup *)
      mutable killed : bool;
      mutable running : (int, task * Unix.inet_addr) Hashtbl.t;
        (* key: request id *)
      mutable clients : MCSet.t;
      mutable status : status;
      esys : Unixqueue.event_system;
      port : int;
      work_dir : string;
      mutable next_req_id : int;
    }


let misalloc_penalty = 2.0
  (** Misallocated tasks have this factor bigger weight. The idea is to
     prevent misallocations as long as good allocations are possible.
     (A misallocated task is run on a non-preferred host.)
   *)

let real_version =
  Mapred_config.executable_version()

let new_client job saddr =
  let mclient_config =
    Rpc_proxy.ManagedClient.create_mclient_config
      ~programs:[ Mapred_rpc_clnt.Mapred_task.V1._program ]
      ~initial_ping:true
      () in
  (* TODO: no timeout so far - this is not possible with current RPC API *)
  let mclient =
    Rpc_proxy.ManagedClient.create_mclient
      mclient_config
      (Plasma_util.connector_of_sockaddr saddr)
      job.esys in
  job.clients <- MCSet.add mclient job.clients;
  mclient


let new_configured_client job saddr onlimit_opt when_ready =
  let mclient = new_client job saddr in
  let pending = ref 0 in
  let check_ready() =
    if !pending = 0 then
      when_ready mclient in
  Mr_proxy.V1.configure'async 
    mclient 
    job.config.res_conf
    (fun get_reply -> 
       ( try let () = get_reply() in ()
	 with error ->
	   let ip = Plasma_util.ip_of_sockaddr saddr in
	   Netlog.logf `Err "Exception from configure to %s: %s"
	     (Unix.string_of_inet_addr ip)
	     (Netexn.to_string error)
       );
       decr pending;
       check_ready());
  incr pending;
  ( match onlimit_opt with
      | None -> ()
      | Some onlimit ->
	  Mr_proxy.V1.job_capacity'async
	    mclient
	    (Mapred_job_config.mapred_job_config job.config.mjc)#job_id
	    (fun get_reply ->
	       ( try
		   let res = get_reply() in
		   onlimit res
		 with error ->
		   let ip = Plasma_util.ip_of_sockaddr saddr in
		   Netlog.logf `Err "Exception from job_capacity to %s: %s"
		     (Unix.string_of_inet_addr ip)
		     (Netexn.to_string error)
	       );
	       decr pending;
	       check_ready()
	    );
	  incr pending;
  )



let dispose_client job c =
  job.clients <- MCSet.remove c job.clients


let kill job =
  Unix.sleep 1;  (* give servers time to log errors *)
  Netlog.logf `Info "Killing job";
  job.killed <- true;
  Hashtbl.iter
    (fun req_id (task,ip) ->
       let task_s = Mapred_tasks.string_of_task_id task in
       Netlog.logf `Info "Submitting kill request of %s to %s"
	 task_s (Unix.string_of_inet_addr ip);
       let saddr = Unix.ADDR_INET(ip,job.port) in
       let c = new_client job saddr in
       Mr_proxy.V1.kill'async
	 c
	 req_id
	 (fun get_reply ->
	    Rpc_proxy.ManagedClient.shut_down c;
	    try
	      let () = get_reply() in
	      Netlog.logf `Info "Kill request for %s was accepted" task_s
	    with
	      | error ->
		  Netlog.logf `Err
		    "Exception after kill request for %s to %s: %s"
		    task_s (Unix.string_of_inet_addr ip)
		    (Netexn.to_string error)
	 )
    )
    job.running;
  Hashtbl.clear job.running;
  if job.status = `Running then
    job.status <- `Killed
    


let start esys plan jc rjc port =
  let load = Hashtbl.create 5 in
  (** (io_load,cpu_load) per IP addr *)

  let load_limit = Hashtbl.create 5 in
  (** dynamic load limit per IP addr *)

  let retry_later = ref TaskSet.empty in
  (** contains tasks that responded recently with `Retry_later. This should
      not be immediately retried, but first when some resources were freed
      again.
   *)

  let retry_later_nodes = Hashtbl.create 5 in
  (** The corresponding full nodes *)

  let plan_dumped = ref false in

  let jc = Mapred_job_config.mapred_job_config rjc.mjc in
  let job =
    { config = rjc;
      plan = plan;
      started_tasks = [];
      killed = false;
      running = Hashtbl.create 7;
      clients = MCSet.empty;
      status = `Running;
      esys = esys;
      port = port;
      work_dir = jc#work_dir;
      next_req_id = 0;
    } in


  let get_load ip =
    try Hashtbl.find load ip with Not_found -> (0.0, 0.0) in

  let update_load_limit ip lim =
    (* lim: number of cores that are assigned for this job. We double the
       number of tasks we send to the machine, just to be sure we max it
       out. Also, add 0.5 to avoid rounding problems, and ensure we can
       at least execute one task.
     *)
    if lim > 0.0 then (
      let lim' =
	max 1.0 (2.0 *. lim) +. 0.5 in
      Hashtbl.replace load_limit ip lim'
    ) in

  let submit task pref_hosts (io_weight,cpu_weight) onsuccess =
    (** Try to start [task] on the [pref_hosts]. If there are no preferences,
       tries to start on any host. Returns [true] if possible, else [false].
     *)
    let hosts =
      if pref_hosts = [] then 
	List.map snd (Mapred_sched.hosts plan)
      else
	pref_hosts in
    if hosts = [] then
      failwith "List of hosts is empty";
    let hosts_sorted =
      List.sort
	(fun ip1 ip2 ->
	   let (io_ld1, cpu_ld1) = get_load ip1 in
	   let (io_ld2, cpu_ld2) = get_load ip2 in
	   compare (io_ld1 +. cpu_ld1) (io_ld2 +. cpu_ld2)
	)
	hosts in
    let ip_opt =
      try
	Some
	  (List.find
	     (fun ip ->
		let limit =
		  try Hashtbl.find load_limit ip
		  with Not_found -> rjc.load_limit in
		let (ip_io_load, ip_cpu_load) = get_load ip in
		ip_io_load +. io_weight <= limit && 
		  ip_cpu_load +. cpu_weight <= limit
	     )
	     hosts_sorted)
      with Not_found -> None in

    let exec ip req_id task_s c =
      Mr_proxy.V1.exec'async
	c
	(req_id, 
	 Mapred_job_config.marshal rjc.mjc, 
	 Mapred_tasks.encode_task task)
	(fun get_reply ->
	   let res =
	     try Mapred_tasks.decode_task_result(get_reply())
	     with error ->
	       Netlog.logf `Err "Exception from exec to %s: %s"
		 (Unix.string_of_inet_addr ip)
		 (Netexn.to_string error);
	       `Error("RPC error: " ^ Netexn.to_string error) in
	   Rpc_proxy.ManagedClient.shut_down c;
	   dispose_client job c;
	   let (old_ip_io_load, old_ip_cpu_load) = get_load ip in
	   let new_ip_io_load = old_ip_io_load -. io_weight in
	   let new_ip_cpu_load = old_ip_cpu_load -. cpu_weight in
	   Hashtbl.replace load ip (new_ip_io_load, new_ip_cpu_load);
	   Hashtbl.remove job.running req_id;
	   match res with
	     | `Ok files ->
		 Netlog.logf `Info "Finished task %s" task_s;
		 ( match task with
			 | `Map mt ->
			     Mapred_sched.add_map_output
			       plan mt files
			 | _ -> ()
		 );
		 if rjc.dump_plan_when_complete && 
		   Mapred_sched.plan_complete plan && not !plan_dumped
		 then (
		   printf "COMPLETE PLAN:\n";
		   Mapred_sched.print_plan plan;
		   flush stdout;
		   plan_dumped := true
		 );
		 Mapred_sched.mark_as_finished plan task;
		 Hashtbl.remove retry_later_nodes ip;
		 retry_later := TaskSet.empty;
		 onsuccess()
	     | `Retry_later ->
		 if Hashtbl.length job.running = 0 then (
		   let msg = 
		     "Got Retry_later code for the only running task" in
		   Netlog.logf `Warning "%s" msg;
		   (* keep retry_later as is - so all other tasks remain
		      blocked
		    *)
		   Hashtbl.clear retry_later_nodes;
		   Mapred_sched.remove_marks plan task;
		   onsuccess()   (* i.e. not a fatal error *)
		 )
		 else (
		   Netlog.logf `Info "Delaying task %s" task_s;
		   retry_later := TaskSet.add task !retry_later;
		   Hashtbl.replace retry_later_nodes ip true;
		   Mapred_sched.remove_marks plan task;
		   onsuccess()   (* i.e. not a fatal error *)
		 )
	     | `Corrupt_input _ ->
		 assert false
	     | `Error msg ->
		 Netlog.logf `Err "Error for task %s: %s" task_s msg;
		 let x_msg = sprintf "Task %s: %s" task_s msg in
		 job.status <- `Errors [ x_msg ];
		 if not job.killed then
		   kill job
	);

 in

    match ip_opt with
      | Some ip ->
	  let (ip_io_load, ip_cpu_load) = get_load ip in
	  let saddr = Unix.ADDR_INET(ip,port) in
	  let onlimit_opt = Some (update_load_limit ip) in
	  let req_id = job.next_req_id in
	  job.next_req_id <- job.next_req_id + 1;
	  let task_s = 
	    sprintf "%s #%06d" (Mapred_tasks.string_of_task_id task) req_id in
	  new_configured_client job saddr onlimit_opt (exec ip req_id task_s);

	  let new_ip_io_load = ip_io_load +. io_weight in
	  let new_ip_cpu_load = ip_cpu_load +. cpu_weight in
	  Hashtbl.replace load ip (new_ip_io_load, new_ip_cpu_load);
	  Hashtbl.replace job.running req_id (task,ip);
	  Mapred_sched.mark_as_started plan task;
	  job.started_tasks <- task :: job.started_tasks;
	  Netlog.logf `Info
	    "Submitted task %s to %s"
	    task_s
	    (Unix.string_of_inet_addr ip);
	  true
      | None ->
	  false
  in

  let rec determine_pref_hosts task =
    (** see comment above *)
    match task with
      | `Map mt -> 
	  mt.map_best_hosts
      | `Sort st ->
	  List.flatten
	    (List.map
	       determine_pref_hosts
	       (Mapred_sched.task_depends_on_list plan task))
      | `Shuffle st ->
	  List.flatten
	    (List.map
	       determine_pref_hosts
	       (Mapred_sched.task_depends_on_list plan task))
      | `Cleanup ->
	  assert false
  in

  let determine_weight task misalloc_flag =
    let factor = 
      if misalloc_flag then misalloc_penalty else 1.0 in
    let (w_io, w_cpu) =
      match task with
	| `Cleanup -> (0.0, 0.0)
	| `Map _ -> rjc.map_weight
	| `Sort _ -> rjc.sort_weight
	| `Shuffle t -> 
	    if t.shuffle_reduce then
	      rjc.reduce_weight
	    else
	      rjc.shuffle_weight in
    (factor *. w_io, w_cpu)
  in

  let node_is_ok ip =
    not (Hashtbl.mem retry_later_nodes ip)
  in

  let rec loop () =
    let tasks0 = Mapred_sched.executable_tasks plan in
    let tasks = 
      List.filter (fun t -> not (TaskSet.mem t !retry_later)) tasks0 in
    Netlog.logf `Info
      "Stats: runnable=%d running=%d finished=%d total=%d complete=%B"
      (List.length tasks)
      (Mapred_sched.n_running plan)
      (Mapred_sched.n_finished plan)
      (Mapred_sched.n_total plan)
      (Mapred_sched.plan_complete plan);
    if not job.killed && tasks <> [] then (
      (** Get the pref_hosts for each task. Also, split into tasks with
	  preference and in tasks without. The tasks with pref will be
	  tried first for getting them allocated
       *)
      let tasks_prefs =
	List.map
	  (fun t ->
	     (t, List.filter node_is_ok (determine_pref_hosts t))
	  )
	  tasks in
      let tasks_with, tasks_without =
	List.partition
	  (fun (t, prefs) -> prefs <> [])
	  tasks_prefs in
      (* TODO (perf): in the following, avoid to check for further job 
	 submissions if all nodes are already maxed out
       *)
      (** First try to start tasks with preferred hosts. If we reach
	  the load limit, tasks may not be startable, so some rest
	  will be left
       *)
      let tasks_with_nonstart =
	List.filter
	  (fun (t, prefs) ->
	     let weight = determine_weight t false in
	     let is_started = submit t prefs weight loop in
	     not is_started
	  )
	  tasks_with in
      (** Now try to start any other tasks *)
      List.iter
	(fun (t, _) ->
	   let weight = determine_weight t false in
	   ignore(submit t [] weight loop)
	)
	tasks_without;
      List.iter
	(fun (t, _) ->
	   let weight = determine_weight t true in
	   ignore(submit t [] weight loop)
	)
	tasks_with_nonstart;
    )
    else
      if (Mapred_sched.plan_finished plan && 
	    Mapred_sched.n_finished plan = Mapred_sched.n_total plan)
      then (
	Netlog.logf `Info "Finished job";
	job.status <- `Successful
      )
  in

  Netlog.logf `Info "Checking executable versions";
  List.iter
    (fun (node,ip) ->
       let saddr = Unix.ADDR_INET(ip,job.port) in
       let client = new_client job saddr in
       let ok = Mr_proxy.V1.check_version client real_version in
       if not ok then
	 failwith ("Node runs wrong task server executable: " ^ node);
       Rpc_proxy.ManagedClient.shut_down client;
       dispose_client job client;
    )
    (Mapred_sched.hosts plan);

  Netlog.logf `Info "Starting job";
  loop();
  job


let status job = job.status


let files_to_keep task =
  (** This only looks at *_del flags *)
  match task with
    | `Map mt ->
	[]
    | `Sort st ->
	if not st.sort_input_del then
	  List.map (fun (n,_,_) -> n) st.sort_input
	else
	  []
    | `Shuffle st ->
	if not st.shuffle_input_del then
	  List.flatten
	    (List.map 
	       (fun (f,_,_) -> 
		  List.map (fun (n,_,_) -> n) f
	       )
	       st.shuffle_input
	    )
	else
	  []
    | `Cleanup -> []

let cleanup job =
  Netlog.logf `Info "Sending cleanup tasks";
  List.iter
    (fun (node,ip) ->
       let saddr = Unix.ADDR_INET(ip,job.port) in
       let c = new_client job saddr in
       let req_id = job.next_req_id in
       job.next_req_id <- job.next_req_id + 1;
       let rs =
	 Mr_proxy.V1.exec
	   c
	   (req_id,
	    Mapred_job_config.marshal job.config.mjc, 
	    Mapred_tasks.encode_task `Cleanup) in
       Rpc_proxy.ManagedClient.shut_down c;
       let r = Mapred_tasks.decode_task_result rs in
       match r with
	 | `Ok _ -> ()
	 | `Retry_later ->
	     Netlog.logf `Err "Got Retry_later from cleanup task"
	 | `Corrupt_input _ ->
	     Netlog.logf `Err "Got Corrupt_input from cleanup task"
	 | `Error msg ->
	     Netlog.logf `Err "Error for task Cleanup: %s" msg
    )
    (Mapred_sched.hosts job.plan);

  Netlog.logf `Info "Cleaning rest up";
  let to_keep = Hashtbl.create 5 in
  List.iter
    (fun task ->
       let keep = files_to_keep task in
       List.iter (fun f -> Hashtbl.replace to_keep f ()) keep;
    )
    job.started_tasks;
  let c = Mapred_sched.cluster job.plan in
  let trans = Plasma_client.start c in
  let to_delete = Plasma_client.list trans job.work_dir in
  Plasma_client.commit trans;

  List.iter
    (fun (filebase,_) ->
       let file = job.work_dir ^ "/" ^ filebase in
       if not (Hashtbl.mem to_keep file) then (
	 let trans = Plasma_client.start c in
	 ignore(Plasma_client.unlink trans file);
	 Plasma_client.commit trans
       )
    )
    to_delete;
  Netlog.logf `Info "Done clean-up"
