(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_commit.ml 332 2011-01-04 01:51:32Z gerd $ *)

open Plasma_util.Operators
open Printf

let dlogf = Plasma_util.dlogf


let slave_info (hp,mc,enabled) =
  let s =
    match Rpc_proxy.ManagedClient.mclient_state mc with
      | `Down -> "down"
      | `Connecting -> "connecting"
      | `Up None -> "up(n/a)"
      | `Up (Some addr) -> "up(" ^ Netsys.string_of_sockaddr addr ^ ")" in
  sprintf "slave(addr=%s,enabled=%B,serial=%d,state=%s)"
    hp !enabled
    (Rpc_proxy.ManagedClient.mclient_serial mc) s


let disable_slave shared_state (hp,mclient,enabled) err =
  Netlog.logf `Alert
    "Disabling namenode %s after error: %s"
    (slave_info (hp,mclient,enabled)) (Netexn.to_string err);
  enabled := false;
  Unix.sleep 2;  (* debug *)
  Nn_push.trigger_shutdown mclient
    (* The shutdown will be considered as rollback by the slave *)



let commit_phase1_write_locally journal shared_state t =
  (** Execute the db updates in the journal *)
  Netlog.logf `Info "Writing changes to local database";
  let esys = shared_state#event_system in
  qseq_e (Queue.copy journal) esys
    (fun m ->
       Nn_db.exec_e m t#connection esys
    )
    
let commit_phase1_prepare_locally t =
  Netlog.logf `Info "Preparing commit at local database";
  t#prepare()


let commit_phase1_write_remotely slave journal shared_state rev_id =
  (** Start remote transactions, and execute the updates from the journal
      there. Errors are suppressed (but logged) - individual slaves may
      fail. Later we check that the number of alive slaves is still at
      least the minimum.
   *)
  let (hp, mclient, enabled) = slave in
  let esys = shared_state#event_system in
  let cname = shared_state # node_config # nn_clustername in
   if !enabled then (
    Netlog.logf `Info "Pushing updates to slave %s" hp;
    (Nn_push.begin_transaction_e mclient cname rev_id)
    ++ (fun _ ->
	  qseq_e (Queue.copy journal) esys
	    (fun m ->
	       if !enabled then
		 Nn_push.push_e m mclient
	       else (
		 Netlog.logf `Warning
		   "commit_phase1_write_remotely still running \
                    although slave disabled";
		 eps_e (`Done ()) esys
	       )
	    )
       )
    >> (function
	  | `Done _  -> `Done ()
	  | `Error e -> disable_slave shared_state slave e; `Done()
	  | `Aborted -> 
	      disable_slave shared_state slave (Failure "Aborted");
	      `Done()
       )
  ) else
    (* Skip by providing a "false positive" return value *)
    eps_e (`Done ()) esys
      

let commit_phase1_prepare_remotely slave shared_state =
  (** Prepare remote transactions for commit. Errors are handled as in
      commit_phase1_write_remotely
   *)
  let (hp, mclient, enabled) = slave in
  let esys = shared_state#event_system in
  if !enabled then (
    Netlog.logf `Info "Preparing commit at slave %s" hp;
    (Nn_push.prepare_commit_e mclient)
    ++ (fun b ->
	  if b then
	    eps_e (`Done()) esys
	  else
	    eps_e (`Error(Failure "Slave denied transaction")) esys
       )
    >> (function
	  | `Done _  -> `Done ()
	  | `Error e -> disable_slave shared_state slave e; `Done()
	  | `Aborted -> 
	      disable_slave shared_state slave (Failure "Aborted");
	      `Done()
       )
  ) else
    (* Skip by providing a "false positive" return value *)
    eps_e (`Done ()) esys


let commit_phase1 journal shared_state t rev_id =
  (** Phase 1: Write changes to database; Start transaction on
      remote hosts; Write changes to remote hosts. Prepare
      the db transactions on remote hosts. Be as parallel
      as possible. 
   *)
  let esys = shared_state#event_system in
  let e_local = 
    commit_phase1_write_locally journal shared_state t in
  let e_remote slave =
    (commit_phase1_write_remotely slave journal shared_state rev_id)
      ++ (fun _ ->
	    commit_phase1_prepare_remotely slave shared_state
	 ) in
  let e_remotes =
    List.map e_remote shared_state#slaves in
  (msync_e (e_local :: e_remotes) esys)
    ++ (fun _ -> 
	  (** Phase 1 end: Prepare the db transaction on the local db.
              Once it is prepared, the transaction will be committed on
              restart after failure!
	      
              We also check here whether there are enough non-disabled
              slaves.
	   *)
	  if shared_state#enough_slaves then
	    commit_phase1_prepare_locally t
	  else (
	    eps_e 
	      (`Error(Failure "Number of slaves drops below required minimum. \
                              Aborting transaction."))
	      esys
	  )
       )


let commit_phase2_commit_remotely shared_state slave =
  (** Commit remote transactions. Errors are handled as in
      commit_phase1_write_remotely
   *)
  let esys = shared_state#event_system in
  let (hp, mclient, enabled) = slave in
  if !enabled then (
    Netlog.logf `Info "Doing commit at slave %s" hp;
    (Nn_push.commit_e mclient)
    >> (function
	  | `Done _ ->
	      Netlog.logf `Info "Successful commit at slave %s" hp;
	      `Done ()
	  | `Error e ->
	      disable_slave shared_state slave e; 
	      `Done()
	  | `Aborted ->
	      disable_slave shared_state slave (Failure "Aborted");
	      `Done()
       )
  ) else
    (* Skip by providing a "false positive" return value *)
    eps_e (`Done ()) esys


let commit_phase2 shared_state dbmods =
  (** Commit all remote transactions, and wait for ACK from all.
      If anything fails, we only disable the slave, but we do not
      roll back anything (which is impossible anyway for more than
      1 slave).
   *)
  let esys = shared_state#event_system in
  msync_e
    ( (List.map
	 (commit_phase2_commit_remotely shared_state) 
	 shared_state#slaves
      )
    )
    esys
    

let commit_e get_journal post_commit (shared_state:Nn_state.shared_state_t) name =
  let esys = shared_state#event_system in
  let pc_e (flag:bool) =
    (** The [post_commit] call must be run inside this prioritizer. *)
    shared_state # blockmap_prioritizer # prioritized
      (fun _ ->
	 post_commit flag;
	 eps_e (`Done()) esys
      )
      1
    ++ (fun () -> eps_e (`Done flag) esys) in

  shared_state # finish_transaction_serialized
    (fun _ ->
       let journal = get_journal() in
       (** Catch simple case: No writes. In this case, just commit the local
	   transaction. The slaves are not contacted at all.
	*)
       if Queue.is_empty journal then
	 pc_e true
       else (
	 (* If we already have too few slaves do not even start: *)
	 if not(shared_state#enough_slaves) then (
	   Netlog.logf `Err
	     "Transaction %s: Aborting commit because the number of slaves is \
              below the minimum" name;
	   pc_e false
	 )
	 else (
	   let t = Nn_db.transact esys in
	   (** Generate new revision id and push the update to journal *)
	   let prev_id = shared_state # cur_rev_ident in
	   let rev_id = shared_state # new_rev_ident() in
	   Queue.push (`Revision_upd rev_id) journal;
	   
	   Netlog.logf `Info "Transaction %s: Starting commit of revision %s" 
	     name rev_id;
	   
	   (** The commit is done in two phases. Phase 1 just writes the
               changes to the databases, and prepares the commit so it
               cannot fail anymore because of SQL errors or insufficient
               disk space. If phase 1 is not successful, all db 
               transactions (local and remote) are rolled back.
	       
               Otherwise we continue with phase 2. The transactions are
               now committed: First all remote transactions, and second -
               after getting ACKs from all slaves - the local transaction.
	       
               We generally implement the policy "once the transaction is
               prepared for commit, it will be committed". This is no problem
               as we have the revision identifier as protection against
               the consequences of losing the synchronization of the 
               databases. The identifier allows us to check that all
               databases are in the same state, and which state is the latest
               one.
	       
               The user will see success if at least the local db can
               finally commit the transaction. If slaves fail in phase 2,
               we just record that, and prevent any further writes.
	    *)
	   let dbmods = 
	     Queue.fold
	       (fun acc m -> m::acc)
	       [] 
	       journal in
	   (
	     (commit_phase1 journal shared_state t prev_id)
	     ++ (fun _ -> commit_phase2 shared_state dbmods)
	     >> (function
		   | `Done _ -> `Done true
		   | `Error e ->
		       Netlog.logf `Err
			 "Transaction %s: \
                          Rolling transaction back after error: %s"
			 name (Netexn.to_string e);
		       `Done false
		   | `Aborted ->
		       Netlog.logf `Err
			 "Transaction %s: Rolling transaction back after \
                          abortion" name;
		       `Done false
		)
	   )
	   ++ (fun commit_flag ->
		 if commit_flag then (
		   (** Rest of phase 2: local commit *)
		   Netlog.logf `Info "Doing commit at local database";
		   (* Any error will disable the master until restart! *)
		   ( ( t#commit()
			 (** Just after the commit we can send the inode
			     notifications. Only when sent at this late point 
			     it is guaranteed that clients get a consistent
			     view: especially that clients see the new data
			     in the db after getting the notification.
			  *)
		       ++ (fun _ -> Nn_notify.send_notifications_e dbmods)
		     )
		     >> (function
			   | `Done _ ->
			       Netlog.logf `Info
				 "Successful commit of revision %s"
				 rev_id;
			       shared_state # set_rev_ident rev_id;
			       `Done true
			   | `Error err ->
			       Netlog.logf `Alert
				 "Cannot commit transaction in master \
                                      db: %s"
				 (Netexn.to_string err);
			       shared_state # master_enabled := false;
			       `Done false
			   | `Aborted -> `Aborted
			)
		   )
		   ++ pc_e
		 )
		 else (
		   Netlog.logf `Info "Rolling back at local database";
		   t#rollback()
		   ++ (fun _ ->
			 Netlog.logf `Info
			   "Aborted commit of revision %s" rev_id;
			 pc_e false
		      )
		 )
	      )
	)
    )
  )


let trans_e shared_state
            (vl:Nn_state.view_t list) 
            (arg_e: 'a Uq_engines.engine) 
            name
	    : _ Uq_engines.final_state Uq_engines.engine =
  (arg_e >>
     (function
	| `Done x -> `Done (`Done x)
	| `Error err -> `Done (`Error err)
	| `Aborted -> `Done `Aborted
     )
  )
  ++ (fun st ->
	let is_ok =
	  match st with
	    | `Done _ -> true
	    | _ -> false in
	let finish_e =
	  if is_ok then (
	   commit_e
	     (fun () ->  (* get_journal *)
		let q = Queue.create() in
		List.iter
		  (fun v ->
		     Queue.transfer v#pre_commit q
		  )
		  vl;
		q
	     )
	     (fun trans_success ->  (* post_commit *)
		List.iter
		  (fun v -> v#post_commit trans_success)
		  vl
	     )
	     shared_state
	     name
	  )
	  else (
	    List.iter
	     (fun v -> v#rollback())
	      vl;
	    eps_e (`Done false) arg_e#event_system
	  ) in
	finish_e
	++ (fun committed ->
	      if committed then
		eps_e (`Done(st)) arg_e#event_system
	      else (
		(** Keep original error if possible *)
		let st' =
		  match st with
		   | `Done _ -> 
		       `Error(Failure "Nn_commit: Commit failed")
		   | _ ->
		       `Done st in
		eps_e st' arg_e#event_system
	      )
	   )
     )
    
    
