(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_config.mli 361 2011-01-31 17:35:47Z gerd $ *)

(** Configuration of name nodes *)

class type nn_node_config =
object
  method nn_clustername : string
    (** The clustername *)
  method nn_hostname : string
    (** The hostname of this node. Must be resolvable to a reasonable
        IP address so that other nodes can reach this node.
     *)
  method nn_nodes : string list
    (** The name nodes in "host:port" syntax *)
  method nn_node_timeout : float
    (** The timeout for connects and RPC calls to name nodes *)
  method nn_node_alive_min : int
    (** How many name nodes need to be alive to continue with operations *)
  method nn_node_alive_min_startup : int
    (** How many name nodes need to be alive at cluster startup *)
  method nn_rank : string
    (** The rank of this node compared to other nodes. At startup time,
        the node with the lowest rank is chosen as coordinator.
        Each node must have a distinct rank. Ranks are compared as in
	[String.compare].
     *)
  method nn_elect_timeout : float
    (** For how many seconds to wait for neighbor namenodes *)
  method nn_inodecaches : string list
    (** Which inodecaches are announced to the clients *)
  method nn_replication : int
    (** Default replication factor *)
  method dn_blocksize : int
    (** Expected blocksize of datanodes *)
  method dn_nodes : string list
    (** The data nodes in "host:port" syntax *)
  method dn_node_timeout : float
    (** The timeout for connects and RPC calls to data nodes *)
end


val extract_node_config : Netplex_types.config_file -> nn_node_config
  (** Config section must look like (see {!Nn_config.nn_node_config}
      for explanations):

      {[
         netplex {
           ...
           namenodes {
             clustername = "cluster1";  (* required *)
             node { addr = "hostname:port" };
             node { addr = "hostname:port" };
             ...
             hostname = "hostname";     (* optional *)
             timeout = 60.0;            (* optional *)
             alive_min = 1;             (* optional *)
             alive_min_startup = 1;     (* optional *)
             elect_timeout 300.0;       (* optional *)
             rank = "0";                (* mandatory *)
             replication = 3;           (* optional *)
             inodecache { port = <port> };
           }
           datanodes {
             node { addr = "hostname:port" }
             node { addr = "hostname:port" }
             ...
             timeout = 60.0;            (* optional *)
             blocksize = 65536;         (* mandatory *)
           }
           ...
         }
      ]}

      The parameter [rank] can also be given by running a script that
      outputs the rank to stdout. This script must be given as 
      parameter [rank_script].

      Instead of describing the nodes directly in the config file
      using [node] subsections, one can also point to external file:

      {[ 
         node_list = "filename";
         port = <default_port>;
      ]}

      These two parameters can replace both the [node] subsections for
      [namenodes] and for [datanodes]. The external file enumerates the
      hosts, one hostname a line (with shell-style comments). 

      Special values:
      - [alive_min] and [alive_min_startup] can also be set to "all"
        (in double quotes). This means all configured name nodes must
        be available
      - [replication] can also be set to "max" (in double quotes). This
        means that the replication is set to the number of data nodes.
   *)

val bmaprowsize : int
  (** block map row size - how many bits are in one row of the block map.
      Changing this value also affects the database.
   *)
