(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_coord.ml 332 2011-01-04 01:51:32Z gerd $ *)

(** The main module for the coordinator (master). Defines RPC services
    for [Filesystem], [Coordination], and [Dn_admin].

    The [Nameslave] program is also "implemented", but all calls 
    (except [null]) fail.
 *)

open Printf
open Plasma_util.Operators

let dlogf = Plasma_util.dlogf

module Rpcapi_clnt = Pfs_rpcapi_clnt
module Rpcapi_srv = Pfs_rpcapi_srv
module Rpcapi_aux = Pfs_rpcapi_aux
module A = Pfs_rpcapi_aux

module Node_config_var =
  Netplex_cenv.Make_var_type(struct type t = Nn_config.nn_node_config end)

module Shared_state_var =
  Netplex_cenv.Make_var_type(struct type t = Nn_state.shared_state_t end)

let cfgname = "current_node_config"
let shstname = "current_shared_state"

let get_shared() =
  Shared_state_var.get shstname


let post_start_hook _ conf index rev =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let sh = new Nn_state.shared_state esys conf index rev in
  Node_config_var.set cfgname conf;
  Shared_state_var.set shstname sh;

  let _tm =
    Netplex_cenv.create_timer
      (fun tm ->
	 ignore(Nn_notify.expire_e());
	 true
      ) 1.0 in
  ()


(**********************************************************************)
(* Dn_admin                                                           *)
(**********************************************************************)

(** Implementation of [Dn_admin]: *)

let proc_add_datanode sess (identity,size) emit =
  (** Adds the [datastore] row to the db *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let sh = get_shared() in
  let v = new Nn_state.ds_view sh in
  let e = 
    Nn_commit.trans_e 
      sh [ (v :> Nn_state.view_t) ]
      (	v#datastore_new_id_e()
	++ (fun id ->
	      v#datastore_upd_e ~id ~identity ~size ~enabled:false
	      ++ (fun _ ->
		    eps_e (`Done id) esys
		 )
	   )
      ) "proc_add_datanode" in
  let emit1 id =
    (** Notify [nn_monitor]: *)
    Nn_datastore_news.signal_db_change id None;
    emit () in
  Plasma_util.rpc_srv_reply "Dn_admin.add_datanode" sess emit1 e


let proc_enable_or_disable procname enable sess (identity,hp) emit =
  (** Modifies the [datastore] db row *)
  if enable then ( (** syntax check only *)
    ignore(Plasma_util.parse_host_port hp)
  );
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let sh = get_shared() in
  let v = new Nn_state.ds_view sh in
  let e = 
    Nn_commit.trans_e
      sh [(v :> Nn_state.view_t)]
      ( v#datastore_list_e()
	++ (fun stores ->
	      try
		let ds =
		  try 
		    List.find 
		      (fun ds -> ds.Nn_datastores.identity = identity) 
		      stores
		  with Not_found -> 
		    failwith (procname ^ ": identity not found") in
		let id = ds.Nn_datastores.id in
		let size = ds.Nn_datastores.size in
		v#datastore_upd_e ~id ~identity ~size ~enabled:enable
		++ (fun _ ->
		      eps_e (`Done id) esys
		   )
	      with
		| error ->
		    eps_e (`Error error) esys
	   )
      )
      "proc_enable_or_disable" in
  let emit1 id =
    (** Notify [nn_monitor]: *)
    let node_opt =
      if enable then
	let (h,p) = Plasma_util.parse_host_port hp in
	let addr = Plasma_util.sockaddr_of_host_port (h,p) in
	Some addr
      else
	None in
    Nn_datastore_news.signal_db_change id node_opt;
    (** Notify Nn_datanode_ctrl: *)
    if not enable then (
      Nn_datanode_ctrl.disconnect id
    );
    emit () in
  Plasma_util.rpc_srv_reply procname sess emit1 e


let proc_enable_datanode =
  proc_enable_or_disable "Dn_admin.enable_datanode" true

let proc_disable_datanode sess identity emit =
  proc_enable_or_disable
    "Dn_admin.disable_datanode" false sess (identity,"") emit

let proc_is_enabled sess (identity) emit =
  (** Get from the [Monitor] service whether this datastore is enabled. *)
  (** First force that our version of the datastore list is updated from
      the [Monitor] process. This is ok here because this is an admin 
      interface, and we want to be as exact as possible
   *)
  Nn_datastore_news.force();
  (** Get the information: *)
  try
    let ds = Nn_datastore_news.get identity in (* or Not_found *)
    emit ds.Nn_datastores.enabled
  with
    | Not_found ->
	Netlog.logf `Err
	  "Dn_admin.is_enabled: Unknown datastore: %s" identity;
	emit false


let proc_lookup sess (identity) emit =
  (** Do the lookup (similar to [proc_is_enabled]): *)
  Nn_datastore_news.force();
  try
    let ds = Nn_datastore_news.get identity in (* or Not_found *)
    let str =
      if ds.Nn_datastores.enabled then
	match ds.Nn_datastores.node with
	  | Some addr -> 
	      ( match addr with
		  | Unix.ADDR_INET(ip,p) ->
		      sprintf "%s:%d"
			(Unix.string_of_inet_addr ip) p
		  | _ ->
		      assert false
	      )
	  | None -> ""
      else
	"" in
    emit str
  with
    | Not_found ->
	Netlog.logf `Err
	  "Dn_admin.is_enabled: Unknown datastore: %s" identity;
	Rpc_server.reply_error sess Rpc.System_err


let proc_datanodes sess () emit =
  (** Do the lookup (similar to [proc_is_enabled]): *)
  Nn_datastore_news.force();
  emit
    (Array.of_list
       (List.map
	  (fun ds -> ds.Nn_datastores.identity)
	  (Nn_datastore_news.list())
       )
    )

let proc_destroy_datanode sess (identity) emit =
  (** Deletes the [datastore] db row *)
  let procname = "destroy_datanode" in
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let sh = get_shared() in
  let v = new Nn_state.ds_view sh  in
  let e = 
    Nn_commit.trans_e
      sh [(v :> Nn_state.view_t)]
      ( v#datastore_list_e()
	++ (fun stores ->
	      try
		let ds =
		  try 
		    List.find 
		      (fun ds -> ds.Nn_datastores.identity = identity) 
		      stores
		  with Not_found -> 
		    failwith (procname ^ ": identity not found") in
		let id = ds.Nn_datastores.id in
		v#datastore_del_e ~id
		++ (fun _ ->
		      eps_e (`Done id) esys
		   )
	      with
		| error ->
		    eps_e (`Error error) esys
	   )
      ) "proc_destroy_datanode" in
  let emit1 id =
    (** Notify [nn_monitor]: *)
    Nn_datastore_news.signal_db_change id None;
    (** Notify Nn_datanode_ctrl: *)
    Nn_datanode_ctrl.disconnect id;
    emit () in
  Plasma_util.rpc_srv_reply procname sess emit1 e


(**********************************************************************)
(* Filesystem                                                         *)
(**********************************************************************)

(** Implementation of [Filesystem]: *)

type fs_transaction =
    { bm_view : Nn_state.bm_view_t;
      inode_view : Nn_state.inode_view_t;
      ticket_verifier : Nn_datanode_ctrl.ticket_verifier;
      mutable running : bool;   (* whether an operation is running *)
      mutable pending_abort : bool;
        (* a pending abort is done first when the current operation is
	   finished
	 *)
    }

type fs_trans_tbl =
    (Rpc_server.connection_id * int64, fs_transaction) Hashtbl.t

module Fs_trans_tbl_var =
  Netplex_cenv.Make_var_type(struct type t = fs_trans_tbl end)

module Int64_var =
  Netplex_cenv.Make_var_type(struct type t = int64 end)

let fs_trans_tbl_name = "current_transtbl"

let get_trans_tbl() =
  try
    Fs_trans_tbl_var.get fs_trans_tbl_name
  with
    | _ ->
	let ht = Hashtbl.create 7 in
	Fs_trans_tbl_var.set fs_trans_tbl_name ht;
	ht

let sess_peer sess =
  try Netsys.string_of_sockaddr
    (Rpc_server.get_peer_name sess)
  with _ -> "n/a"

let new_transaction_e sess tid =
  let sh = get_shared() in
  try
    let cid = Rpc_server.get_connection_id sess in
    let transtbl = get_trans_tbl() in
    let bm_v = new Nn_state.bm_view sh in
    let inode_v = new Nn_state.inode_view sh in
    let fstr =
      { bm_view = bm_v;
	inode_view = inode_v;
	ticket_verifier = Nn_datanode_ctrl.new_ticket_verifier();
	running = false;
	pending_abort = false;
      } in
    Hashtbl.replace transtbl (cid,tid) fstr;
    dlogf
      "New transaction %Ld (peer %s)"
      tid (sess_peer sess);
    eps_e (`Done fstr) sh#event_system
  with error ->
    eps_e (`Error error) sh#event_system


let del_transaction cid tid =
  let transtbl = get_trans_tbl() in
  Hashtbl.remove transtbl (cid,tid)

let fs_rpc_srv_reply procname sess emit te =
  (** Handle a server reply. When a [Transaction_error] is found, set
      [errno] to this exception. For unknown exceptions, set [errno] to
      [`efailed], the generic error.

      The engine [te] must follow the conventions for transactional
      engines. Such engines never transition to [`Error] or [`Abort],
      but always to [`Done] (corresponding to the fact that transactions
      handle errors so they do not propagate to the "outside"). The [`Done]
      state has the original state value as argument, i.e. we see
      [`Done(`Done v)], [`Done(`Error e)], and [`Done `Aborted], so
      the value of the inner engine enclosed by the transaction handler is
      still visible.
   *)
  let te' =
    te 
    >> (function
	  | `Done(`Done x) ->
	      `Done(`Done(`ok x :> [Nn_state.errno | `ok of _] ))
	  | `Done(`Error(Nn_state.Transaction_error code as err)) ->
	      Netlog.logf `Err
		"Error in %s: %s" procname (Netexn.to_string err);
	      `Done(`Done(code :> [Nn_state.errno | `ok of _] ))
	  | `Done(`Error err) ->
	      Netlog.logf `Err
		"Error in %s: %s" procname (Netexn.to_string err);
	      `Done(`Done `efailed)
	  | `Done `Aborted ->
	      `Done `Aborted
	  | `Done (`Working _) ->
	      assert false
	  | `Error err ->
	      `Error err
	  | `Aborted ->
	      `Aborted
       ) in
  Plasma_util.rpc_srv_reply
    procname sess emit te'

let fs_rpc_srv_reply_unit ?(anyway=fun()->()) procname sess emit te =
  (** A variant of the former for unit/void replies *)
  let te' =
    te 
    >> (function
	  | `Done(`Done ()) ->
	      `Done(`Done(`ok :> [Nn_state.errno | `ok ] ))
	  | `Done(`Error(Nn_state.Transaction_error code as err)) ->
	      Netlog.logf `Err
		"Error in %s: %s" procname (Netexn.to_string err);
	      `Done(`Done(code :> [Nn_state.errno | `ok ] ))
	  | `Done(`Error err) ->
	      Netlog.logf `Err
		"Error in %s: %s" procname (Netexn.to_string err);
	      `Done(`Done `efailed)
	  | `Done `Aborted ->
	      `Done `Aborted
	  | `Done (`Working _) ->
	      assert false
	  | `Error err ->
	      `Error err
	  | `Aborted ->
	      `Aborted
       ) in
  Plasma_util.rpc_srv_reply
    ~onerror:(fun _ -> anyway())
    ~onaborted:(fun _ -> anyway())
    ~ondone:(fun _ -> anyway())
    procname sess emit te'


let abort_trans cid tid trans =
  del_transaction cid tid;
  trans.bm_view # rollback();
  trans.inode_view # rollback();
  ignore(Nn_datanode_ctrl.disable_ticket_verifier_e
	   trans.ticket_verifier)
    


let with_transaction_1 rpc_srv_reply procname sess emit tid f =
  (** Ensure [f] is run within a transaction. If the transaction is already
      started everything is fine. 

     [sess], [emit]: from the RPC. [procname] is the name of the RPC
     for messages
   *)

  let esys = (Netplex_cenv.self_cont()) # event_system in
  let transtbl = get_trans_tbl() in
  let cid = Rpc_server.get_connection_id sess in
  let trans_opt =
    try Some(Hashtbl.find transtbl (cid,tid))
    with Not_found -> None in
  let run_f (g : _ Uq_engines.engine -> _ Uq_engines.engine) trans =
    trans.running <- true;
    let te = 
      try
	(** [g] is for mapping results so the result is always [`Done].
            Also for commit/abort in autocommit mode.
	 *)
	g(f trans) 
      with
	| error ->
	    eps_e (`Done (`Error error)) esys in
    let te' =
      te
      ++ (fun r ->
	    trans.running <- false;
	    if trans.pending_abort then (
	      Netlog.logf `Info
		"Executing pending rollback of zombie transaction %Ld" tid;
	      abort_trans cid tid trans;
	      trans.pending_abort <- false;
	      eps_e (`Done (`Error (Nn_state.Transaction_error `efailed))) esys
	    )
	    else 
	      eps_e (`Done r) esys
	 ) in
    rpc_srv_reply procname sess emit te'
  in
  match trans_opt with
    | Some trans ->
	(** The transaction is already started. *)
	if trans.running then
	  emit `etbusy
	else
	  run_f 
	    (fun e1 ->
	       map_e
		 ~map_done:(fun v -> `Done(`Done v))
		 ~map_error:(fun e -> `Done(`Error e))
		 ~map_aborted:(fun _ ->`Done `Aborted)
		 e1
	    )
	    trans
    | None -> (
	(** There is no transaction yet *)
	dlogf
	  "ENOTRANS for transaction %Ld (peer %s)"
	  tid (sess_peer sess);
	emit `enotrans
      )

let with_transaction procname =
  with_transaction_1 fs_rpc_srv_reply procname

let with_transaction_unit =
  with_transaction_1 fs_rpc_srv_reply_unit

let fake_trans e =
  (** Wrap [`Done] in another [`Done], so it looks like as if the value
      came from a transactional engine
   *)
  map_e
    ~map_done:(fun v -> `Done(`Done v))
    e

let trans_error errno =
  Nn_state.Transaction_error errno


let blockmap_dealloc_e ~inode bmv blist =
  let bm_dealloc =
    Nn_blocklist.fold
      (fun identity index_range bi acc ->
	 let tuple =
	   ( bi.Rpcapi_aux.identity, 
	     bi.Rpcapi_aux.block,
	     bi.Rpcapi_aux.length ) in
	 tuple :: acc
      )
      blist
      [] in
  let total =
    List.fold_left
      (fun acc (_,_,l) -> Int64.add l acc) 0L bm_dealloc in
  dlogf
    "blockmap_dealloc-e: #blocks=%Ld" total;
  bmv # blockmap_dealloc_e ~inode bm_dealloc


let delete_inode_e trans tid id =
  (** Sequence of actions:
      1. Get inodeinfo, and blocklimit
      2. Get list of blocks [old_bis] in the FULL range (can be long)
      3. Deallocate the [old_bis] blocks in the blockmap
      4. Remove inode
   *)
  (* TODO: for very large files it can be a problem that step 2 returns
     the complete list. This is not fixable by performing 2./3. in a 
     loop, because the blocks are also cached. Maybe delete the cache
     when it becomes too full? 
   *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  trans.inode_view # inode_get_with_lock_e ~id
    (* lock needed by inode_dealloc_e anyway *)
  ++ (fun ii_opt ->
	match ii_opt with
	  | None ->
	      eps_e (`Error(Nn_state.Transaction_error `estale)) esys
	  | Some ii ->
	      dlogf
		"delete_inode: found inode";
	      eps_e (`Done ii) esys)
  ++ (fun ii ->
	trans.inode_view # inode_get_blocks_e
	  ~id ~blkidx:0L ~len:ii.A.blocklimit ~keep_flag:false trans.bm_view)
  ++ (fun old_bis ->
	blockmap_dealloc_e ~inode:id trans.bm_view old_bis)
  ++ (fun () ->
	trans.inode_view # inode_dealloc_e ~id
     )


let delete_inodes_e trans tid idlist =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let q = Queue.create() in
  List.iter (fun id -> Queue.add id q) idlist;
  Plasma_util.q_seq_engine q esys
    (fun id -> delete_inode_e trans tid id)

    
    
let proc_begin_transaction sess tid emit =
  (** When a transaction is started, the datanodes are notified of it
      (the [Nn_datanode_ctrl] call below), so they can accept block writes
      for this transaction.
   *)
  let transtbl = get_trans_tbl() in
  let cid = Rpc_server.get_connection_id sess in
  let trans_opt =
    try
      Some(Hashtbl.find transtbl (cid,tid))
    with Not_found -> None in
  match trans_opt with
    | None ->
	fs_rpc_srv_reply_unit
	  "begin_transaction"
	  sess emit 
	  (fake_trans
	     (new_transaction_e sess tid
	      ++ (fun trans ->
		    Nn_datanode_ctrl.enable_ticket_verifier_e
		      trans.ticket_verifier
		 )
	     )
	  )
    | Some _ ->
	Netlog.logf `Err
	  "Error: begin_transaction but transaction %Ld already active" tid;
	emit `efailed

let proc_commit_transaction sess tid emit =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let transtbl = get_trans_tbl() in
  let cid = Rpc_server.get_connection_id sess in
  let trans_opt =
    try
      Some(Hashtbl.find transtbl (cid,tid))
    with Not_found -> None in
  match trans_opt with
    | None ->
	Netlog.logf `Err
	  "Error: commit_transaction outside transaction";
	emit `enotrans

    | Some trans ->
	if trans.running then
	  emit `etbusy
	else (
	  trans.running <- true;
	  let sh = get_shared() in
	  fs_rpc_srv_reply_unit
	    ~anyway:(fun () -> 
		       trans.running <- false;
		       if trans.pending_abort then (
			 let exists = Hashtbl.mem transtbl (cid,tid) in
			 if exists then
			   abort_trans cid tid trans;
			 Netlog.logf `Info
			   "Zombie transaction %Ld has been %s" 
			   tid
			   (if exists then "rolled back" else "committed");
			 trans.pending_abort <- false;
		       )
		    )
	    "commit_transaction"
	    sess 
	    emit
	    (map_e
	       ~map_done:(fun (_,commit_result) -> `Done commit_result)
	       (new Uq_engines.sync_engine
		  (** On commit we do two things in parallel: 
		      1. Notify the datanodes (so block writes become
                         impossible again), 
		      2. execute the commit protocol.
		   *)
		  (Nn_datanode_ctrl.disable_ticket_verifier_e 
		     trans.ticket_verifier)
		  (map_e
		     ~map_done:(fun v -> 
				  del_transaction cid tid;
				  dlogf
				    "Committed transaction %Ld (peer %s)"
				    tid (sess_peer sess);
				  `Done v
			       )
		     (delete_inodes_e 
			trans tid
			trans.inode_view#delayed_inode_deletes
		      ++ (fun () ->
			    Nn_commit.trans_e
			      sh 
			      [ (trans.bm_view :> Nn_state.view_t); 
				(trans.inode_view :> Nn_state.view_t) ]
			      (eps_e (`Done()) esys)
			      (sprintf "%Ld (peer %s)"
				 tid (sess_peer sess))
			 )
		     )
		  )
	       )
	    )
	)

let proc_abort_transaction sess tid emit =
  let transtbl = get_trans_tbl() in
  let cid = Rpc_server.get_connection_id sess in
  let trans_opt =
    try
      Some(Hashtbl.find transtbl (cid,tid))
    with Not_found -> None in
  match trans_opt with
    | None ->
	Netlog.logf `Err "Error: abort_transaction outside transaction";
	emit `enotrans;

    | Some trans ->
	if trans.running then
	  (* FIXME: better we could allow this *)
	  emit `etbusy
	else (
	  dlogf
	    "Aborted transaction %Ld (peer %s)"
	    tid (sess_peer sess);
	  trans.running <- true;
	  del_transaction cid tid;
	  trans.bm_view # rollback();
	  trans.inode_view # rollback();
	  
	  fs_rpc_srv_reply_unit
	    "abort_transaction"
	    sess 
	    (fun r -> 
	       trans.running <- false; 
	       if trans.pending_abort then (
		 Netlog.logf `Info
		   "Zombie transaction %Ld has been rolled back" tid;
		 trans.pending_abort <- false
	       );
	       emit r)
	    (fake_trans
	       (Nn_datanode_ctrl.disable_ticket_verifier_e 
		  trans.ticket_verifier)
	    )
	)


let fs_onclose srv cid =
  (** If the TCP connection is closed, implicitly roll all transactions back *)
  let transtbl = get_trans_tbl() in
  let trans_l =
    Hashtbl.fold 
      (fun (cid1,tid1) wtrans acc ->
	 if cid1 = cid then
	   (tid1,wtrans) :: acc
	 else
	   acc
      )
      transtbl
      [] in
  List.iter
    (fun (tid,trans) ->
       let peer =
	 try Netsys.string_of_sockaddr
	   (Rpc_server.get_conn_peer_name cid)
	 with _ -> "n/a" in
       if trans.running then (
	 if not trans.pending_abort then (
	   Netlog.logf `Info
	     "Scheduling termination of zombie transaction %Ld (peer %s)"
	     tid peer;
	   trans.pending_abort <- true
	 )
       )
       else (
	 Netlog.logf `Info
	   "Rolling transaction %Ld back after TCP close (peer %s)" 
	   tid peer;
	 abort_trans cid tid trans
       )
    )
    trans_l

(** The following RPC implementations are just transaction-wrapped
    incovations of the right methods in [Nn_state].
 *)

let fixup_ii ii =
  let module A = Pfs_rpcapi_aux in
  let sh = get_shared() in
  if ii.A.mtime.A.tnsecs < 0 || ii.A.mtime.A.tnsecs >= 1_000_000_000 then
    raise(Nn_state.Transaction_error `einval);
  if ii.A.ctime.A.tnsecs < 0 || ii.A.ctime.A.tnsecs >= 1_000_000_000 then
    raise(Nn_state.Transaction_error `einval);
  if ii.A.replication < 0 then
    raise(Nn_state.Transaction_error `einval);
  if ii.A.eof < 0L then
    raise(Nn_state.Transaction_error `einval);
  let t0_lz = lazy(Unix.gettimeofday()) in
  if ii.A.mtime.A.tsecs < 0L then
    ii.A.mtime <- Nn_state.encap_time (Lazy.force t0_lz);
  if ii.A.ctime.A.tsecs < 0L then
    ii.A.ctime <- Nn_state.encap_time (Lazy.force t0_lz);
  if ii.A.replication = 0 then
    ii.A.replication <- sh#node_config#nn_replication

let proc_allocate_inode sess (tid,ii) emit =
  let module A = Pfs_rpcapi_aux in
  with_transaction 
    "allocate_inode" sess emit tid
    (fun trans ->
       fixup_ii ii;
       trans.inode_view # inode_alloc_e ii
    )

let proc_delete_inode sess (tid,id) emit =
  with_transaction_unit
    "delete_inode" sess emit tid
    (fun trans ->
       delete_inode_e trans tid id
    )

let proc_update_inodeinfo sess (tid,id,ii) emit =
  with_transaction_unit
    "update_inodeinfo" sess emit tid
    (fun trans ->
       fixup_ii ii;
       trans.inode_view # inode_upd_e ~id ~keep_blocklimit:true ii
    )

let proc_get_inodeinfo sess (tid,id) emit =
  with_transaction 
    "get_inodeinfo" sess emit tid
    (fun trans ->
       new Uq_engines.map_engine
	 ~map_done:(function
		      | None ->
			  `Error(Nn_state.Transaction_error `estale)
		      | Some ii ->
			  `Done ii
		   )
	 (trans.inode_view # inode_get_e ~id)
    )

let proc_get_blocks sess (tid,inode,blkidx,len,keep_flag) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "get_blocks" sess emit tid
    (fun trans ->
       if len < 0L || blkidx > Int64.sub Int64.max_int len then
	 raise(Nn_state.Transaction_error `einval);
       trans.inode_view # inode_get_e 
	 ~id:inode
       ++ (fun ii_opt ->
	     match ii_opt with
	       | None ->
		   eps_e (`Error(Nn_state.Transaction_error `estale)) esys
	       | Some _ ->
		   (* don't look at this ii - it may already be outdated! *)
		   eps_e (`Done ()) esys
	  )
       ++ (fun () ->
	     trans.inode_view # inode_get_blocks_e 
	       ~id:inode ~blkidx ~len ~keep_flag trans.bm_view
	  )
       ++ (fun bis ->
	     (* finally set node_alive *)
	     eps_e
	       (let bis' =
		  Array.map
		    (fun bi -> 
		       let node_alive =
			 Nn_datastore_news.is_alive ~wait:true
			   bi.Rpcapi_aux.identity in
		       { bi with 
			   Rpcapi_aux.node_alive = node_alive
		       }
		    )
		    (Array.of_list
		       (Nn_blocklist.to_blockinfo_list bis)) in
		`Done bis'
	       )
	       esys
	  )
    )

let proc_allocate_blocks sess (tid,inode,blkidx,len,set_mtime,pref) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "allocate_blocks" sess emit tid
    (fun trans ->
       if len < 0L || blkidx > Int64.sub Int64.max_int len then
	 raise(Nn_state.Transaction_error `einval);
       if len > Int64.of_int max_int then
	 raise(Nn_state.Transaction_error `einval);
       let tv = trans.ticket_verifier in
       let td =
	 { Nn_alloc_types.td_idlist = tv.Nn_datanode_ctrl.tv_datanodes;
	   td_safetrans_id = tv.Nn_datanode_ctrl.tv_safetrans_id;
	   td_safetrans_secret = tv.Nn_datanode_ctrl.tv_safetrans_sec;
	   td_safetrans_tmo = tv.Nn_datanode_ctrl.tv_safetrans_tmo;
	 } in
       (** Sequence of actions:
	   1. Get the inode (needed for repfactor)
	   2. Get list of blocks [old_bis] in the range (blkidx,len)
           3. Deallocate the [old_bis] blocks in the blockmap
           4. Allocate new blocks [new_bis] from the blockmap
           5. Remove [old_bis] from the inode list, add [new_bis]
	   6. Update inode if blocklimit has changed or [set_mtime]

           We need to lock early in step 1. Otherwise bad things can happen:
            - A competing transaction could also allocate blocks in the same
	      range, and if the transaction is committed between steps 2 and 3
	      we would deallocate the wrong blocklist
	    - A competing transaction could change the inode, and we
	      return the wrong seqno's. This confuses caches on the client
	      side.
	*)
       trans.inode_view # inode_get_with_lock_e ~id:inode
       ++ (fun ii_opt ->
	     match ii_opt with
	       | None ->
		   eps_e (`Error(Nn_state.Transaction_error `estale)) esys
	       | Some ii ->
		   dlogf
		     "allocate_blocks: found inode";
		   eps_e (`Done ii) esys)
	 ++ (fun ii ->
	       trans.inode_view # inode_get_blocks_e
		 ~id:inode ~blkidx ~len ~keep_flag:false trans.bm_view
	       ++ (fun old_bis ->
		     blockmap_dealloc_e ~inode trans.bm_view old_bis
		  )
	       ++ (fun () ->
		     dlogf
		       "allocate_blocks: old blocks deallocated";
		     trans.inode_view # allocinfo_get_e ~id:inode
		  )
	       ++ (fun ai_blocks ->
		     dlogf
		       "allocate_blocks: got allocinfo";
		     let ri =
		       { Nn_blockmap.ri_blocks = ai_blocks
		       } in
		     trans.bm_view # blockmap_alloc_e 
		       ~inode
		       ~rpolicy:(`Balanced ii.Rpcapi_aux.replication)
		       ~index:blkidx
		       ~number:(Int64.to_int len)
		       ~pref:(Array.to_list pref)
		       td ri
		  )
	       ++ (fun new_bis ->
		     dlogf
		       "allocate_blocks: new blocks allocated";
		     let new_limit =
		       max 
			 ii.Rpcapi_aux.blocklimit
			 (Int64.add blkidx len) in
		     trans.inode_view # inode_upd_blocks_e 
		       ~id:inode
		       ~del:(blkidx, len)
		       ~add:new_bis
		       ~set_blocklimit:new_limit
		       ~set_mtime
		     ++ (fun _ ->
			   dlogf
			     "allocate_blocks: new blocks inserted";
			   let blocks_l =
			     Nn_blocklist.to_blockinfo_list new_bis in
			   let blocks_a =
			     Array.map
			       (fun bi ->
				  { bi with A.inode_seqno = ii.A.seqno } )
			       (Array.of_list blocks_l) in
			   eps_e (`Done blocks_a) esys
			)
		  )
	    )
    )


let proc_free_blocks sess (tid,inode,blkidx,len,set_mtime) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "free_blocks" sess emit tid
    (fun trans ->
       if len < 0L || blkidx > Int64.sub Int64.max_int len then
	 raise(Nn_state.Transaction_error `einval);
       (** Sequence of actions:
	   1. Get list of blocks [old_bis] in the range (blkidx,len)
           2. Deallocate the [old_bis] blocks in the blockmap
           3. Remove [old_bis] from the inode list
	*)
       trans.inode_view # inode_get_with_lock_e ~id:inode
       ++ (fun ii_opt ->
	     match ii_opt with
	       | None ->
		   eps_e (`Error(Nn_state.Transaction_error `estale)) esys
	       | Some ii ->
		   eps_e (`Done ii) esys
	  )
       ++ (fun ii ->
	     trans.inode_view # inode_get_blocks_e
	       ~id:inode ~blkidx ~len ~keep_flag:false trans.bm_view)
       ++ (fun old_bis ->
	     blockmap_dealloc_e ~inode trans.bm_view old_bis)
       ++ (fun () ->
	     trans.inode_view # inode_upd_blocks_e 
	       ~id:inode
	       ~del:(blkidx, len)
	       ~add:Nn_blocklist.empty
	       ~set_blocklimit:(-1L)
	       ~set_mtime
	  )
    )
  

let proc_get_fsstat sess _ emit =
  let sh = get_shared() in
  let esys = (Netplex_cenv.self_cont()) # event_system in

  let ds_list = Nn_datastore_news.list ~wait:true () in
  let ds_enabled = 
    List.filter (fun ds -> ds.Nn_datastores.enabled) ds_list in
  let idlist =
    List.map (fun ds -> ds.Nn_datastores.identity) ds_enabled in
  
  let total = ref 0L in
  let used = ref 0L in
  let trans = ref 0L in

  let bm_view = new Nn_state.bm_view sh in
  let e =
    Uq_engines.msync_engine
      (List.map (fun id -> bm_view # blockmap_get_e ~identity:id) idlist)
      (fun bm () ->
	 if bm#active then (
	   total := Int64.add !total bm#size;
	   used := Int64.add !used bm#n_used;
	   trans := Int64.add !trans bm#n_trans
	 )       
      )
      ()
      esys in
  fs_rpc_srv_reply
    "get_fsstat"
    sess emit 
    (fake_trans
       (e ++
	  (fun () -> 
	     let r =
	       { Pfs_rpcapi_aux.total_blocks = !total;
		 used_blocks = !used;
		 trans_blocks = !trans;
		 
		 have_block_checksums = false;
		 have_safetrans_for_reads = false;
		 have_protected_inodes = false
	       } in
	     
	     eps_e (`Done r) esys
	  )))


let proc_rereplicate sess _ emit = assert false


let slash_re = Pcre.regexp "[/]+"

let split_path ?(strip_trailing_slashes=false) p =
  (* empty string: []
     "/": [""]
     absolute path: first element of list is ""
   *)
  if p = "" then
    []
  else
    let max =
      if strip_trailing_slashes then 0 else (-1) in
    let l =
      Pcre.split ~rex:slash_re ~max p in
    if l = [] then  (* "/" *)
      [ "" ]
    else
      l


let rec lookup_path_e trans inode ii path loops symbolic =
  (* walks along relative path starting at inode *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  match path with
    | [] ->
	eps_e (`Done(inode, ii)) esys
    | fn :: path' ->
	if ii.Pfs_rpcapi_aux.filetype <> `ftype_directory then
	  eps_e (`Error (trans_error `ebadpath)) esys
	else (
	  match fn with
	    | "." | "" ->
		(* stay in the same directory *)
		lookup_path_e trans inode ii path' loops symbolic
	    | ".." ->
		(* get the parent directory *)
		trans.inode_view # filename_parent_dir_e ~inode
		++ (fun parent_inode_opt ->
		      match parent_inode_opt with
			| None ->
			    eps_e (`Error (trans_error `enoent)) esys
			| Some parent_inode ->
			    trans.inode_view # inode_get_e ~id:parent_inode
			    ++ (fun parent_ii_opt ->
				  match parent_ii_opt with
				    | None ->
					eps_e 
					  (`Error (trans_error `enoent)) esys
				    | Some parent_ii ->
					lookup_path_e 
					  trans parent_inode parent_ii
					  path' loops symbolic
			       )
		   )
	    | _ ->
		(* lookup the path component *)
		trans.inode_view # filename_get_e ~dir_inode:inode ~name:fn
		++ (fun sub_inode_opt ->
		      match sub_inode_opt with
			| None ->
			    eps_e (`Error (trans_error `enoent)) esys
			| Some sub_inode ->
			    trans.inode_view # inode_get_e ~id:sub_inode
			    ++ (fun sub_ii_opt ->
				  match sub_ii_opt with
				    | None ->
					eps_e
					  (`Error (trans_error `enoent)) esys
				    | Some sub_ii ->
					(match sub_ii.Pfs_rpcapi_aux.filetype with
					   | `ftype_directory | `ftype_regular->
					       lookup_path_e
						 trans sub_inode sub_ii
						 path' loops symbolic
					   | `ftype_symlink ->
					       if symbolic && path' = [] then
						 eps_e
						   (`Done(sub_inode, sub_ii))
						   esys
					       else
						 lookup_symlink_e
						   trans inode ii
						   sub_inode sub_ii path'
						   loops symbolic
					)
			       )
		   )
	)

and lookup_abs_e trans path loops symbolic =
  (* "/" has the fixed inode 1 *)
  trans.inode_view # inode_get_e ~id:1L
  ++ (fun root_ii_opt ->
	match root_ii_opt with
	  | None ->
	      assert false
	  | Some root_ii ->
	      lookup_path_e trans 1L root_ii path loops symbolic 
     )

and lookup_symlink_e trans inode ii sym_inode sym_ii path' loops symbolic =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let target =
    split_path sym_ii.Pfs_rpcapi_aux.field1 in
  if target = [] then
    eps_e (`Error (trans_error `enoent)) esys
  else
    if !loops >= 100 then
      eps_e (`Error (trans_error `eloop)) esys
    else (
      incr loops;
      if List.hd target = "" then
	lookup_abs_e trans (List.tl target @ path') loops symbolic
      else
	lookup_path_e trans inode ii (target @ path') loops symbolic 
    )
	
let lookup_e trans inode pathname symbolic =
  (* returns (inode,ii), or ENOENT *)
  (* inode is only used if pathname is non-absolute *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let path = split_path pathname in
  match path with
    | "" :: path' ->
	lookup_abs_e trans path' (ref 0) symbolic
    | _ ->
	trans.inode_view # inode_get_e ~id:inode
	++ (fun ii_opt ->
	      match ii_opt with
		| None ->
		    eps_e (`Error (trans_error `enoent)) esys
		| Some ii ->
		    lookup_path_e trans inode ii path (ref 0) symbolic
	   )

let no_slash_slash n =
  not(Pcre.pmatch ~pat:"//" n)

let no_trailing_slash n =
  n = "" || n = "/" || n.[ String.length n - 1 ] <> '/'

let check_name n =
  (** Check whether [n] is a good filename *)
  n <> "" && n.[0] = '/' && no_slash_slash n && no_trailing_slash n

let get_parent n =
  (** Return the parent name *)
  try
    if n = "/" then raise Not_found;
    let p = String.rindex n '/' in
    if p=0 then "/" else String.sub n 0 p
  with
    | Not_found ->
	failwith "Nn_coord.get_parent"


let proc_fs_lookup sess (tid,dir_inode,path,symbolic) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "lookup" sess emit tid
    (fun trans ->
       lookup_e trans dir_inode path symbolic
       ++ (fun (inode, ii) ->
	     eps_e (`Done inode) esys
	  )
    )


let rec rev_lookup_e trans dir_inode name acc =
  (* Translates the file identified by the dir_inode and the name member
     into an absolute path. Call with acc=[]
   *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  if dir_inode = 1L then (
    let path =
      "/" ^ String.concat "/" (name :: acc) in
    eps_e (`Done (Some path)) esys
  )
  else
    trans.inode_view # filename_rev_get_e ~inode:dir_inode
    ++ (fun dir_names ->
	  match dir_names with
	    | [] ->
		eps_e (`Done None) esys
	    | [dir_dir_inode, dir_name] ->
		rev_lookup_e trans dir_dir_inode dir_name (name :: acc)
	    | _ ->
		eps_e (`Error (trans_error `efailed)) esys
       )



let proc_rev_lookup sess (tid,inode) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "rev_lookup" sess emit tid
    (fun trans ->
       if inode=1L then
	 eps_e (`Done [| "/" |]) esys
       else
	 trans.inode_view # filename_rev_get_e ~inode
	 ++ (fun spots ->
	       Uq_engines.msync_engine
		 (List.map
		    (fun (dir_inode,name) ->
		       rev_lookup_e trans dir_inode name []
		    )
		    spots)
		 (fun path_opt acc ->
		    match path_opt with
		      | None -> acc
		      | Some path -> path :: acc )
		 []
		 esys
	       ++ (fun paths ->
		     eps_e (`Done(Array.of_list paths)) esys
		  )
	    )
    )


let proc_rev_lookup_dir sess (tid,inode) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "rev_lookup_dir" sess emit tid
    (fun trans ->
       if inode=1L then
	 eps_e (`Done "/" ) esys
       else
	 trans.inode_view # filename_rev_get_dir_e ~inode
	 ++ (fun (dir_inode,name,_) ->
	       rev_lookup_e trans dir_inode name []
	       ++ (fun path_opt ->
		     match path_opt with
		       | None ->
			   eps_e (`Error (trans_error `enoent)) esys
		       | Some path ->
			   eps_e (`Done path) esys
		  )
	    )
    )
    

let proc_link_count sess (tid,inode) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "link_count" sess emit tid
    (fun trans ->
       trans.inode_view # filename_count_e ~inode
       ++ (fun n ->
	     eps_e (`Done n) esys
	  )
    )


let proc_list sess (tid,inode) emit =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction 
    "list" sess emit tid
    (fun trans ->
       trans.inode_view # inode_get_e ~id:inode
       ++ (fun ii_opt ->
	     match ii_opt with
	       | None ->
		   eps_e (`Error(Nn_state.Transaction_error `estale)) esys
	       | Some ii ->
		   let ft = ii.Pfs_rpcapi_aux.filetype in
		   match ft with
		     | `ftype_regular | `ftype_symlink ->
			 eps_e (`Error(Nn_state.Transaction_error `enotdir)) esys
		     | `ftype_directory ->
			 trans.inode_view # filename_list_e ~dir_inode:inode
			 ++ (fun entries ->
			       let rentries =
				 List.map
				   (fun (n,i) ->
				      { Pfs_rpcapi_aux.entry_name = n;
					entry_inode = i
				      }
				   )
				   entries in
			       eps_e (`Done (Array.of_list rentries)) esys
			    )
	  )
    )


let update_mtime trans inode =
  (** Update the mtime of the parent by adding a special
      db instruction: (This avoids that we have to lock the
      parent inode)
   *)
  let now = Unix.gettimeofday() in
  trans.inode_view # inode_upd_time
    ~id:inode
    ~mtime:(Some(Nn_state.encap_time now))
    ~ctime:None


let proc_link_at sess (tid,dir_inode,name,inode) emit =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "link_at" sess emit tid
    (fun trans ->
       if String.contains name '/' then
	 raise(Nn_state.Transaction_error `einval);
       trans.inode_view # filename_link_e ~dir_inode ~name ~inode
       ++ (fun () ->
	     update_mtime trans dir_inode;
	     eps_e (`Done ()) esys
	  )
    )


let rec split_last_off_r path =
  match path with
    | [] ->
	assert false
    | [last] ->
	[], last
    | x :: path' ->
	let (p,last) = split_last_off_r path' in
	(x :: p), last

let split_last_off path =
  (* For an absolute path <> "/" this function returns (dir,name) where
     name is the last component, and dir is the preceding part of the path.
     Raises Transaction_error in case of errors.
   *)
  match path with
    | [] ->
	raise(Nn_state.Transaction_error `einval)
    | "" :: _ :: _ ->
	split_last_off_r path
    | _ ->
	raise(Nn_state.Transaction_error `einval)


let proc_link sess (tid,path,inode) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "link" sess emit tid
    (fun trans ->
       (* FIXME: Be more POSIX-compatible: allow the trailing slash when
	  linking a directory
	*)
       let p = split_path path in
       let dir_p, name = split_last_off p in
       if name = "" then raise(Nn_state.Transaction_error `einval);
       let dir_path =
	 if dir_p = [ "" ] then "/" else String.concat "/" dir_p in
       lookup_e trans (-1L) dir_path false
       ++ (fun (dir_inode, _) ->
	     trans.inode_view # filename_link_e ~dir_inode ~name ~inode
	     ++ (fun () ->
		   update_mtime trans dir_inode;
		   eps_e (`Done ()) esys
		)
	  )
    )


let proc_unlink_at sess (tid, dir_inode, name) emit =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "unlink_at" sess emit tid
    (fun trans ->
       trans.inode_view # filename_unlink_e ~dir_inode ~name
       ++ (fun () ->
	     update_mtime trans dir_inode;
	     eps_e (`Done ()) esys
	  )
    )


let proc_unlink sess (tid,path) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "unlink" sess emit tid
    (fun trans ->
       (* FIXME: Be more POSIX-compatible: allow the trailing slash when
	  unlinking a directory
	*)
       let p = split_path path in
       let dir_p, name = split_last_off p in
       if name = "" then raise(Nn_state.Transaction_error `einval);
       let dir_path =
	 if dir_p = [ "" ] then "/" else String.concat "/" dir_p in
       lookup_e trans (-1L) dir_path false
       ++ (fun (dir_inode, _) ->
	     trans.inode_view # filename_unlink_e ~dir_inode ~name
	     ++ (fun () ->
		   update_mtime trans dir_inode;
		   eps_e (`Done ()) esys
		)
	  )
    )


let proc_rename_at sess (tid,o_dir_inode,o_name,n_dir_inode,n_name) emit =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "rename_at" sess emit tid
    (fun trans ->
       if String.contains n_name '/' then
	 raise(Nn_state.Transaction_error `einval);
       trans.inode_view # filename_rename_e
	 ~from_dir_inode:o_dir_inode
	 ~from_name:o_name
	 ~to_dir_inode:n_dir_inode
	 ~to_name:n_name
       ++ (fun () ->
	     update_mtime trans o_dir_inode;
	     update_mtime trans n_dir_inode;
	     eps_e (`Done ()) esys
	  )
    )


let proc_rename sess (tid,o_path,n_path) emit = 
  let esys = (Netplex_cenv.self_cont()) # event_system in
  with_transaction_unit
    "rename" sess emit tid
    (fun trans ->
       (* FIXME: Be more POSIX-compatible: allow the trailing slash when
	  renaming a directory (in o_path only)
	*)
       let o_p = split_path o_path in
       let o_dir_p, o_name = split_last_off o_p in
       if o_name = "" then raise(Nn_state.Transaction_error `einval);
       let o_dir_path =
	 if o_dir_p = [ "" ] then "/" else String.concat "/" o_dir_p in
       let n_p = split_path n_path in
       let n_dir_p, n_name = split_last_off n_p in
       if n_name = "" then raise(Nn_state.Transaction_error `einval);
       let n_dir_path =
	 if n_dir_p = [ "" ] then "/" else String.concat "/" n_dir_p in
       lookup_e trans (-1L) o_dir_path false
       ++ (fun (o_dir_inode, _) ->
	     lookup_e trans (-1L) n_dir_path false
	     ++ (fun (n_dir_inode, _) ->
		   trans.inode_view # filename_rename_e
		     ~from_dir_inode:o_dir_inode
		     ~from_name:o_name
		     ~to_dir_inode:n_dir_inode
		     ~to_name:n_name
		   ++ (fun () ->
			 update_mtime trans o_dir_inode;
			 update_mtime trans n_dir_inode;
			 eps_e (`Done ()) esys
		      )
		)
	  )
    )


let proc_namelock sess (tid,dir_inode,name) emit =
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let sh = get_shared() in
  with_transaction_unit
    "namelock" sess emit tid
    (fun trans ->
       if String.contains name '/' then
	 raise(Nn_state.Transaction_error `einval);
       let owner = (trans.inode_view :> < > )  in
       match sh # lock_filename (dir_inode,name) `Exists owner  with
	 | Some unlock ->
	     let e = 
	       trans.inode_view # filename_get_e ~dir_inode ~name
	       ++ (fun inode_opt ->
		     match inode_opt with
		       | None -> eps_e (`Error (trans_error `enoent)) esys
		       | Some inode -> eps_e (`Done ()) esys
		  ) in
	     Uq_engines.when_state
	       ~is_error:(fun _ -> unlock())
	       ~is_aborted:(fun _ -> unlock())
	       e;
	     e
	 | None ->
	     raise(Nn_state.Transaction_error `econflict)
    )

    
let proc_set_block_checksum sess _ emit = assert false
let proc_iterate sess _ emit = assert false

(**********************************************************************)
(* Setup                                                              *)
(**********************************************************************)

(** Binding RPC servers: *)

let setup srv nn_conf =
  (** Bind the RPC procedure implementations. Note that this is called once
      outside the container context to discover the bound RPC programs. We
      can ignore this here because this discovery is only meaningful for
      portmapper registrations.
   *)
  try
    let conf = Node_config_var.get cfgname in
    Nn_db.ds_cache_init `From_datastore_news_channel;
    Rpcapi_srv.Coordination.V1.bind_async
      ~proc_null:(fun sess () emit -> emit ())
      ~proc_find_coordinator:(fun sess cname emit -> 
				if cname=conf#nn_clustername then
				  let sh = get_shared() in
				  emit(Some sh#master)
				else
				  emit None)
      ~proc_find_namenodes:(fun sess cname emit ->
			      if cname=conf#nn_clustername then
				emit(Array.of_list conf#nn_nodes)
			      else
				emit [| |]
			   )
      ~proc_is_coordinator:(fun sess cname emit -> 
			      if cname=conf#nn_clustername then
				emit(Nn_elect.get_role() = `Coordinator)
			      else
				emit false
			   )
      ~proc_clustername:(fun sess () emit -> emit conf#nn_clustername)
      ~proc_find_inodecaches:(fun sess cname emit ->
				if cname=conf#nn_clustername then
				  emit(Array.of_list conf#nn_inodecaches)
				else
				  emit [| |]
			     )
      ~proc_local_identities:(fun sess cname emit ->
				try
				  if cname <> conf#nn_clustername then
				    raise Not_found;
				  let addr = Rpc_server.get_peer_name sess in
				  let ip = Plasma_util.ip_of_sockaddr addr in
				  let dslist =
				    Nn_datastore_news.list ~wait:true () in
				  let locallist =
				    List.filter
				      (fun ds ->
					 match ds.Nn_datastores.node with
					   | None -> false
					   | Some addr -> 
					       ip = 
					         Plasma_util.ip_of_sockaddr addr
				      )
				      dslist in
				  let identities =
				    List.map 
				      (fun ds -> ds.Nn_datastores.identity)
				      locallist in
				  emit(Array.of_list identities)
				with
				  | _ ->
				      emit [| |]
			     )
      srv;
    
    Rpcapi_srv.Dn_admin.V1.bind_async
      ~proc_null:(fun _ _ emit -> emit())
      ~proc_add_datanode
      ~proc_enable_datanode
      ~proc_disable_datanode
      ~proc_is_enabled
      ~proc_lookup
      ~proc_datanodes
      ~proc_destroy_datanode
      srv;

    Rpcapi_srv.Filesystem.V1.bind_async
      ~proc_null:(fun _ _ emit -> emit())
      ~proc_begin_transaction
      ~proc_commit_transaction
      ~proc_abort_transaction
      ~proc_get_inodeinfo
      ~proc_allocate_inode
      ~proc_update_inodeinfo
      ~proc_delete_inode
      ~proc_get_blocks
      ~proc_allocate_blocks
      ~proc_free_blocks
      ~proc_get_fsstat
      ~proc_rereplicate
      ~proc_lookup:(proc_fs_lookup)
      ~proc_rev_lookup
      ~proc_rev_lookup_dir
      ~proc_link_count
      ~proc_link
      ~proc_link_at
      ~proc_unlink
      ~proc_unlink_at
      ~proc_list
      ~proc_rename
      ~proc_rename_at
      ~proc_namelock
      ~proc_set_block_checksum
      ~proc_iterate
      ~proc_get_blocksize:(fun _ _ emit -> emit nn_conf#dn_blocksize)
      srv;
    Rpc_server.set_onclose_action srv (fs_onclose srv);
    
    let no_nameslave name _ =
      failwith ("Nameslave." ^ name ^ ": Not available on coordinator node") in
    
    Rpcapi_srv.Nameslave.V1.bind
      ~proc_null:(fun () -> ())
      ~proc_begin_transaction:(no_nameslave "begin_transaction")
      ~proc_prepare_commit:(no_nameslave "prepare_commit")
      ~proc_commit:(no_nameslave "commit")
      ~proc_set_ds_cache:(no_nameslave "set_ds_cache")
      ~proc_push_datastore_upd:(no_nameslave "push_datastore_upd")
      ~proc_push_datastore_del:(no_nameslave "push_datastore_del")
      ~proc_push_revision_upd:(no_nameslave "push_revision_upd")
      ~proc_push_blockalloc_upd:(no_nameslave "push_blockalloc_upd")
      ~proc_push_inode_ins:(no_nameslave "push_inode_ins")
      ~proc_push_inode_upd:(no_nameslave "push_inode_upd")
      ~proc_push_inode_upd_time:(no_nameslave "push_inode_upd_time")
      ~proc_push_inode_del:(no_nameslave "push_inode_del")
      ~proc_push_inodeblocks_ins:(no_nameslave "push_inodeblocks_ins")
      ~proc_push_inodeblocks_del:(no_nameslave "push_inodeblocks_del")
      ~proc_push_names_ins:(no_nameslave "push_names_ins")
      ~proc_push_names_del:(no_nameslave "push_names_del")
      srv;

    Nn_notify.setup srv;  (* also bind Request_notifications *)
  with
    | Netplex_cenv.Not_in_container_thread 
    | Netplex_cenv.Container_variable_not_found _ ->
	(** [setup] was run outside the container context *)
	()
