(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_monitor.ml 332 2011-01-04 01:51:32Z gerd $ *)

open Nn_datastores
open Plasma_util.Operators

let dlogf = Plasma_util.dlogf

module Bool_var =
  Netplex_cenv.Make_var_type(struct type t = bool end)

module Node_config_var =
  Netplex_cenv.Make_var_type(struct type t = Nn_config.nn_node_config end)

let dsname = "current_datastores"
let pubname = "current_is_publishing"
let cfgname = "current_node_config"

let is_publishing() =
  try Bool_var.get pubname
  with _ -> false

let list() =
  Datastores_var.get dsname

let get idstr =
  List.find 
    (fun ds -> ds.identity = idstr)
    (list())

let set_list l =
  (** Put the value into the container variable [dsname]: *)
  Datastores_var.set dsname l;
  if is_publishing() then (
    (** If we are publishing, put the value also in the shared Netplex
       variable [shdsname], and notify the receivers
     *)
    Datastores_shvar.set shdsname l;
    (** The receiver of this message is {!Nn_datastore_news.receive_message} *)
    Netplex_cenv.send_message "Nn_*" shdsname [| "have_monitor_news" |];
    Netlog.logf `Info "Notifying news channel about status update (n=%d)"
      (List.length l);
  )

let add ds =
  let l = list() in
  let p_conflict =
    List.exists
      (fun ds1 -> ds1.id = ds.id || ds1.identity = ds.identity)
      l in
  if p_conflict then
    failwith "Nn_monitor.add: conflicting IDs";
  set_list (ds :: l)
  
let delete idstr =
  let l = list() in
  let l' =
    List.filter (fun ds -> ds.identity <> idstr) l in
  set_list l'

let delete_by_id id =
  let l = list() in
  let l' =
    List.filter (fun ds -> ds.id <> id) l in
  set_list l'

let update f idstr =
  (** Updates the list of datastores by applying [f] to the element [idstr] *)
  let l = list() in
  let found = ref false in
  let l' =
    List.map
      (fun ds -> 
	 if ds.identity = idstr then (
	   found := true;
	   f ds
	 )
	 else
	   ds
      )
      l in
  if !found then
    set_list l'
  else
    failwith ("Nn_monitor: identity not found: " ^ idstr)
  
let enable idstr =
  update
    (fun ds -> { ds with enabled = true })
    idstr

let disable idstr =
  update
    (fun ds -> { ds with enabled = false })
    idstr

let update_from_db ?(when_done = fun () -> ()) id =
  (** Start a new db transaction, and read the table of datastores. Look for
      the row for [id], and update the in-memory version of this row. We have
      to be careful to always finish the transaction (rolling it back on
      error). Call [when_done] when the in-memory version is changed.
   *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let t = Nn_db.transact esys in
  let e = Nn_db.datastore_list_e t#connection esys in
  Uq_engines.when_state
    ~is_done:(fun l ->
		( try
		    let ds =
		      List.find (fun ds1 -> ds1.id = id) l in (*or Not_found *)
		    try
		      update
			(fun ds1 ->
			   { ds1 with
			       size = ds.size;
			       enabled = ds.enabled;
			       node = 
			       ( if not ds.enabled then None else ds1.node );
			   })
			ds.identity
		    with
		      | Failure _ -> (** identity not found  *)
			add ds
		  with
		    | Not_found -> (** If not found in db, delete the row *)
			delete_by_id id
		);
		dlogf "update_from_db: return";
		t#return();
		dlogf "update_from_db: when_done";
		when_done();
	     )
    ~is_error:(fun error ->
		 Netlog.logf `Err
		   "Nn_monitor.update_from_db: exception %s"
		   (Netexn.to_string error);
		 t#return();
	      )
    ~is_aborted:(fun () ->
		   Netlog.logf `Err
		     "Nn_monitor.update_from_db: aborted";
		   t#return();
		)
    e


let init_from_db ?(when_done = fun () -> ()) () =
  (** Start a new transaction, read the table of datastores, and set the
      in-memory version of it to this list.
       Call [when_done] when the in-memory version is changed.
   *)
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let t = Nn_db.transact esys in
  let e = Nn_db.datastore_list_e t#connection esys in
  Uq_engines.when_state
    ~is_done:(fun l ->
		set_list l;
		t # return();
		when_done();
	     )
    ~is_error:(fun error ->
		 Netlog.logf `Err
		   "Nn_monitor.init_from_db: exception %s"
		   (Netexn.to_string error);
		 t # return();
	      )
    ~is_aborted:(fun () ->
		   Netlog.logf `Err
		     "Nn_monitor.init_from_db: aborted";
		   t # return();
		)
    e


let is_alive idstr = true
  (* TODO *)

let start_publishing() =
  if not(is_publishing()) then (
    if Netplex_sharedvar.create_var ~enc:true shdsname then (
      Bool_var.set pubname true;
      set_list (list());
      Netlog.logf `Info "Publishing news about datanode status";
    )
    else
      Netlog.logf `Err "Cannot create shared variable: %s" shdsname
  )


let discover ?(when_done = fun _ -> ()) saddr =
  (** Check what the datanode is reporting about itself, and if it is good
      set the [node] component of the [datastore] to the socket address
   *)
  (* TODO: also include the node into monitoring *)
  let conf = Node_config_var.get cfgname in
  let blocksize = conf#dn_blocksize in
  let l = list() in
  let (h,p) =
    Plasma_util.host_port_of_sockaddr saddr in
  dlogf "discover at %s:%d" h p;
  let conn =
    Plasma_util.connector_of_sockaddr saddr in
  let esys = (Netplex_cenv.self_cont()) # event_system in
  let client =
    Pfs_rpcapi_clnt.Datanode.V1.create_client2
      ~esys
      (`Socket(Rpc.Tcp, conn, Rpc_client.default_socket_config)) in
  Rpc_client.configure client 0 conf#dn_node_timeout;

  let report_error n e =
    Netlog.logf `Err
      "Error in RPC call %s @ %s: %s"
      n
      (Netsys.string_of_sockaddr saddr)
      (Netexn.to_string e);
    Rpc_client.shut_down client;
    when_done None
  in

  try
    (** First, get the node's identity, and look this up in our table: *)
    Pfs_rpcapi_clnt.Datanode.V1.identity'async
      client
      conf#nn_clustername
      (fun get_reply ->
	 try
	   let id = get_reply() in
	   ( try
	       let ds = List.find (fun ds -> ds.identity = id) l in
	       (** ok, we know this identity. Now check whether size and
                   blocksize are compatible
		*)
	       Pfs_rpcapi_clnt.Datanode.V1.blocksize'async
		 client
		 ()
		 (fun get_reply ->
		    try
		      let bs = get_reply() in
		      if bs <> blocksize then
			failwith "Node has wrong blocksize";
		      Pfs_rpcapi_clnt.Datanode.V1.size'async
			client
			()
			(fun get_reply ->
			   try
			     let size = get_reply() in
			     if size < ds.size then
			       failwith "Node has fewer blocks than expected";
			     Rpc_client.shut_down client;
			     update 
			       (fun ds -> { ds with node = Some saddr })
			       id;
			     Netlog.logf `Info
			       "Discovered datanode %s at %s:%d with size %Ld %s"
			       id h p size 
			       (if ds.enabled then "(enabled)" 
			        else "disabled");
			     when_done 
			       (Some { d_identity = id;
				       d_size = size
				     } 
			       )
			   with
			     | error ->
				 report_error "Datanode.V1.size" error
			)
		    with
		      | error ->
			  report_error "Datanode.V1.blocksize" error
		 )
	     with
	       | Not_found -> 
		   Netlog.logf `Warning "Node has unknown identity: %s"
		     (Netsys.string_of_sockaddr saddr);
		   when_done None
	   )
	 with
	   | error ->
	       report_error "Datanode.V1.identity" error
      )
  with
    | error ->
	report_error "Datanode.V1.identity" error


let post_add_hook _ ctrl =
  ctrl # add_plugin Netplex_sharedvar.plugin

let receive_message cont msg args =
  if msg = shdsname then (
    match args with
      | [| "have_db_news"; id_s |] ->
	  Netlog.logf `Info "Got notification about db update";
	  let id = int_of_string id_s in
	  update_from_db id
      | [| "have_db_news_discover"; id_s; hp |] ->
	  Netlog.logf `Info 
	    "Got notification about db update; starting discovery at %s" hp;
	  let id = int_of_string id_s in
	  let (h,p) = Plasma_util.parse_host_port hp in
	  let saddr = Unix.ADDR_INET(Unix.inet_addr_of_string h, p) in
	  update_from_db
	    ~when_done:(fun () -> discover saddr)
	    id
      | _ -> 
	  ()
  )

let post_start_hook _ conf =
  Node_config_var.set cfgname conf;
  Netlog.logf `Info "Reading datastore table";
  init_from_db
    ~when_done:(fun () ->
		  Netlog.logf `Info "Discovering datanodes";
		  let n = ref 0 in
		  List.iter
		    (fun hp ->
		       let (h,p) = Plasma_util.parse_host_port hp in
		       let saddr = 
			 Plasma_util.sockaddr_of_host_port (h,p) in
		       discover
			 ~when_done:(fun _ ->
				       decr n;
				       if !n = 0 then (
					 Netlog.logf `Info 
					   "Discovery done";
					 start_publishing()
				       )
				    )
			 saddr;
		       incr n
		    )
		    conf # dn_nodes;
		  if !n = 0 then (
		    Netlog.logf `Info "Nothing to discover";
		    start_publishing()
		  )
	       )
    ()
